#include "robot.hpp"
#include "ILocationProvider.hpp"
#include "SimpleLocTracker.hpp"
#include "utils.hpp"

#include <SFML/System/Vector2.hpp>
#include <cmath>
#include <json/value.h>


sf::Vector2f move_from(sf::Vector2f from, float rot, float distance) {
    float x = std::cos(rot) * distance;
    float y = std::sin(rot) * distance;
    return sf::Vector2f(from.x + x, from.y + y);
}


int distanceToEncodingCount(float distance_mm) {
    const float distance_per_encoded = 10.21;
    return int(distance_mm / distance_per_encoded);
}


int angleToEncodingCount(float angle_rad) {
    const float rad_per_encoded = 0.157;
    return int(angle_rad / rad_per_encoded);
}


Robot::Robot(const std::string & hostIpAddress, uint16_t tcpPort, ILocationProvider* loc_tracker)
    : _jsonRpcTcpClient(hostIpAddress, tcpPort)
    , _irProximitysDistanceDetected(this)
    , _lineTracksIsDetected(this, 1, false)
    , _lineTracksValue(this, 1, 0)
    , _encoderWheelsValue(this, 2, 0)
    , _switchsIsDetected(this)
    , _ultrasoundsDistanceDetected(this)
    , _isReadySemaphore(0)
    , _eventCvMutex()
    , _eventCv()
    , _loc_tracker(loc_tracker)
{
    _jsonRpcTcpClient.bindNotification("irProximityDistanceDetected", [this](const Json::Value & params){
        std::size_t index = params["index"].asUInt();
        _irProximitysDistanceDetected.set(index, params["value"].asUInt(), params["changedCount"].asInt());
    });
    _jsonRpcTcpClient.bindNotification("lineTrackIsDetected", [this](const Json::Value & params){
        std::size_t index = params["index"].asUInt();
        _lineTracksIsDetected.set(index, params["value"].asBool(), params["changedCount"].asInt());
    });
    _jsonRpcTcpClient.bindNotification("lineTrackValue", [this](const Json::Value & params){
        std::size_t index = params["index"].asUInt();
        _lineTracksValue.set(index, params["value"].asUInt(), params["changedCount"].asInt());
    });
    _jsonRpcTcpClient.bindNotification("encoderWheelValue", [this](const Json::Value & params){
        // this->_loc_tracker->progress();
        std::size_t index = params["index"].asUInt();
        auto steps = params["changedCount"].asInt();
        _encoderWheelsValue.set(index, params["value"].asUInt(), steps);
    });
    _jsonRpcTcpClient.bindNotification("switchIsDetected", [this](const Json::Value & params){
        std::size_t index = params["index"].asUInt();
        _switchsIsDetected.set(index, params["value"].asBool(), params["changedCount"].asInt());
    });
    _jsonRpcTcpClient.bindNotification("ultrasoundDistanceDetected", [this](const Json::Value & params){
        std::size_t index = params["index"].asUInt();
        _ultrasoundsDistanceDetected.set(index, params["value"].asUInt(), params["changedCount"].asInt());
    });
    _jsonRpcTcpClient.bindNotification("setIsReady", [this](const Json::Value & params){
        assert(params.isNull());
        _isReadySemaphore.release();
    });
    _jsonRpcTcpClient.startReceive();
}

Robot::~Robot()
{}

void Robot::setMotorPower(MotorIndex motorIndex, float value)
{
    Json::Value params;
    params["motorIndex"] = motorIndexToStringHelper(motorIndex);
    params["value"] = value;
    _jsonRpcTcpClient.callNotification("setMotorPower", params);

}

void Robot::setMotorsPower(float rightValue, float leftValue)
{
    Json::Value params;
    params["rightValue"] = rightValue;
    params["leftValue"] = leftValue;
    _jsonRpcTcpClient.callNotification("setMotorsPower", params);
}

void Robot::doStraightLine(float distance_mm, float power) {
    Json::Value params;
    auto encodingSteps = distanceToEncodingCount(distance_mm);
    params["steps"] = encodingSteps;
    params["power"] = power;
    // DBG(_loc_tracker->get_pos().x);
    // // DBG(_loc_tracker->get_pos().y);
    // auto pos = move_from(_loc_tracker->get_pos(), _loc_tracker->get_rot(), distance_mm);
    // // DBG(pos.x);
    // // DBG(pos.y);
    // _loc_tracker->start(pos, _loc_tracker->get_rot(), encodingSteps * 2);
    _jsonRpcTcpClient.callNotification("straightLine", params);
    DBG("doStraightLine");
    DBG(encodingSteps);
    _loc_tracker->hint_left(true);
    _loc_tracker->hint_right(true);
}

void Robot::doRotate(float angle_rad, float power) {
    Json::Value params;
    auto encodingSteps = angleToEncodingCount(angle_rad);
    params["steps"] = encodingSteps;
    params["power"] = power;
    // _loc_tracker->start(_loc_tracker->get_pos(), _loc_tracker->get_rot() + angle_rad, encodingSteps * 2);
    _jsonRpcTcpClient.callNotification("rotate", params);
    DBG("doRotate");
    DBG(encodingSteps);
    _loc_tracker->hint_left(false);
    _loc_tracker->hint_right(true);
}

void Robot::waitChanged(EventType eventType) {
    auto eventTypes = waitParamHelper({eventType});
    waitChangedHelper(eventTypes);
}

EventType Robot::waitChanged(const std::set<EventType> & eventTypes)
{
    auto eventTypesWithChangedCount = waitParamHelper(eventTypes);
    return waitChangedHelper(eventTypesWithChangedCount);
}

void Robot::notify(EventType eventType, int changedCount)
{
    {
        std::lock_guard<std::mutex> lk(_eventCvMutex);
        _lastNotifiedEventType.insert_or_assign(eventType, changedCount);
    }
    _eventCv.notify_all();
}

std::string Robot::motorIndexToStringHelper(MotorIndex motorIndex)
{
    switch (motorIndex)
    {
        case MotorIndex::RIGHT:
            return "RIGHT";
        case MotorIndex::LEFT:
            return "LEFT";
    }
    throw std::invalid_argument(std::string("Cannot convert ")
            + std::to_string(static_cast<int>(motorIndex)) + " into Robot::MotorIndex");
}

void Robot::waitChangedHelper(EventType eventType, int & changedCount) {
    std::map<EventType, int> eventTypes{{eventType, changedCount}};
    waitChangedHelper(eventTypes);
    changedCount = eventTypes.at(eventType);
}

EventType Robot::waitChangedHelper(std::map<EventType, int> & eventTypes) {
    std::unique_lock<std::mutex> lk(_eventCvMutex);
    EventType notifiedEventType;
    _eventCv.wait(lk, [eventTypes, &notifiedEventType, lastNotifiedEventType = std::ref(_lastNotifiedEventType)]{
        for (auto eventType : eventTypes) {
            if (lastNotifiedEventType.get().at(eventType.first)>eventType.second)
            {
                lastNotifiedEventType.get().at(eventType.first) = eventType.second;
                notifiedEventType = eventType.first;
                return true;
            }
        }
        return false;
    });
    return notifiedEventType;
}

std::map<EventType, int> Robot::waitParamHelper(std::set<EventType> eventTypes)
{
    std::map<EventType, int> toReturn;
    std::unique_lock<std::mutex> lk(_eventCvMutex);
    for (auto eventType : eventTypes)
    {
        int changedCount = -1;
        auto it = _lastNotifiedEventType.find(eventType);
        if (it != _lastNotifiedEventType.end())
            changedCount = it->second;
        toReturn.insert(std::make_pair(eventType, changedCount));
    }
    return toReturn;
}
